using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Sprites;

/*
 * This program demonstrates lighting a 2D scene in XNA
 * Reach model using normal maps.
 * 
 * by Brian MacIntosh
 * 4/12/12
 * 
 * You are free to use and implement this code, explicitly the
 * code in this file (Game1.cs), the shader (lighting.fx) and
 * sprite libraries (SpriteSheet.cs and the SpriteSheetPipeline
 * project), in your own projects.
 * 
 * You may NOT use any of the art (.PNG) resources provided
 * for this demonstration.
 */

namespace SpriteNormals
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        //Resources for drawing the animated sprite
        private SpriteSheet sheet;
        private Sprite sprite;

        //Resources for drawing the background
        private Texture2D bg;
        private Texture2D bgnorm;

        //Resources for drawing the door
        private Texture2D door;
        private Texture2D doorNorm;
        private Texture2D doorSpec;

        //The lighting shader
        private Effect shader;

        //Render targets to
        private RenderTarget2D renderTarget;
        private RenderTarget2D normalRenderTarget;
        private RenderTarget2D specularRenderTarget;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            this.graphics.PreferredBackBufferWidth = 800;
            this.graphics.PreferredBackBufferHeight = 480;

            //this.graphics.IsFullScreen = true;

            this.IsMouseVisible = true;
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            //Load and create a sprite
            sheet = Content.Load<SpriteSheet>("sprites_frames");
            sheet.LoadContent(Content);
            sprite = new Sprite(sheet);
            sprite.PlayAnimation("running");
            sprite.position = new Vector2(300, 300);
            sprite.Loop();

            //Load the background resources
            bg = Content.Load<Texture2D>("background_middle");
            bgnorm = Content.Load<Texture2D>("background_middle_norm");

            //Load the door resources
            door = Content.Load<Texture2D>("door");
            doorNorm = Content.Load<Texture2D>("door_norm");
            doorSpec = Content.Load<Texture2D>("door_spec");

            //Load the shader
            shader = Content.Load<Effect>("lighting");

            //Create render targets for the lighting
            renderTarget = new RenderTarget2D(
                GraphicsDevice,
                GraphicsDevice.PresentationParameters.BackBufferWidth,
                GraphicsDevice.PresentationParameters.BackBufferHeight);
            normalRenderTarget = new RenderTarget2D(
                GraphicsDevice,
                GraphicsDevice.PresentationParameters.BackBufferWidth,
                GraphicsDevice.PresentationParameters.BackBufferHeight);
            specularRenderTarget = new RenderTarget2D(
                GraphicsDevice,
                GraphicsDevice.PresentationParameters.BackBufferWidth,
                GraphicsDevice.PresentationParameters.BackBufferHeight);
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            base.UnloadContent();
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed
                || Keyboard.GetState().IsKeyDown(Keys.Escape))
                this.Exit();

            //Animate the sprite
            sprite.Update(gameTime);

            base.Update(gameTime);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            //Draw initial scene
            GraphicsDevice.SetRenderTarget(renderTarget);
            GraphicsDevice.Clear(Color.Black);

            spriteBatch.Begin(
                SpriteSortMode.Deferred,
                BlendState.NonPremultiplied,
                SamplerState.PointClamp,
                null,
                null);
            spriteBatch.Draw(
                bg,
                Vector2.Zero,
                null,
                Color.White,
                0f,
                Vector2.Zero,
                2f,
                SpriteEffects.None,
                1f);
            spriteBatch.Draw(
                door,
                new Vector2(300, 200),
                null,
                Color.White,
                0f,
                Vector2.Zero,
                2f,
                SpriteEffects.None,
                1f);
            sprite.Draw(
                spriteBatch,
                Color.White,
                0f,
                Vector2.Zero,
                3f);
            spriteBatch.End();

            //Draw normals
            GraphicsDevice.SetRenderTarget(normalRenderTarget);
            GraphicsDevice.Clear(new Color(127, 127, 255));

            spriteBatch.Begin(
                SpriteSortMode.Deferred,
                BlendState.NonPremultiplied,
                SamplerState.PointClamp,
                null,
                null);
            spriteBatch.Draw(
                bgnorm,
                Vector2.Zero,
                null,
                Color.White,
                0f,
                Vector2.Zero,
                2f,
                SpriteEffects.None,
                1f);
            spriteBatch.Draw(
                doorNorm,
                new Vector2(300, 200),
                null,
                Color.White,
                0f,
                Vector2.Zero,
                2f,
                SpriteEffects.None,
                1f);
            sprite.DrawNormal(
                spriteBatch,
                Color.White,
                0f,
                Vector2.Zero,
                3f);
            spriteBatch.End();

            //Draw specular
            GraphicsDevice.SetRenderTarget(specularRenderTarget);
            GraphicsDevice.Clear(Color.Black);

            spriteBatch.Begin(
                SpriteSortMode.Deferred,
                BlendState.NonPremultiplied,
                SamplerState.PointClamp,
                null,
                null);
            spriteBatch.Draw(
                doorSpec,
                new Vector2(300, 200),
                null,
                Color.White,
                0f,
                Vector2.Zero,
                2f,
                SpriteEffects.None,
                1f);
            spriteBatch.End();

            //Draw postprocessed image
            GraphicsDevice.SetRenderTarget(null);

            Vector3 lightPosition = new Vector3(Mouse.GetState().X, Mouse.GetState().Y, 200);
            Vector4 lightColor = new Vector4(0.8f, 1f, 1f, 0);

            shader.Parameters["xNormalMap"].SetValue((Texture2D)normalRenderTarget);
            shader.Parameters["xSpecularMap"].SetValue((Texture2D)specularRenderTarget);
            shader.Parameters["xLightPosition"].SetValue(lightPosition);
            shader.Parameters["xLightColor"].SetValue(lightColor);
            shader.Parameters["xWidth"].SetValue(renderTarget.Width);
            shader.Parameters["xHeight"].SetValue(renderTarget.Height);

            spriteBatch.Begin(
                SpriteSortMode.Immediate,
                BlendState.Opaque,
                SamplerState.LinearClamp,
                null,
                null,
                shader);
            spriteBatch.Draw((Texture2D)renderTarget, Vector2.Zero, Color.White);
            spriteBatch.End();

            base.Draw(gameTime);
        }
    }
}
