﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

namespace Sprites
{
    /// <summary>
    /// This class updates the animations on each sprite each loop
    /// </summary>
    public static class SpriteManager
    {
        /// <summary>
        /// All active sprites in the game
        /// </summary>
        private static List<Sprite> sprites;


        /// <summary>
        /// Create a new sprite manager.
        /// </summary>
        static SpriteManager()
        {
            sprites = new List<Sprite>();
        }


        /// <summary>
        /// Add a new sprite to the game
        /// </summary>
        /// <param name="type"></param>
        /// <returns></returns>
        public static void AddSprite(Sprite sprite)
        {
            sprites.Add(sprite);
        }


        /// <summary>
        /// Remove a sprite from the game
        /// </summary>
        /// <param name="sprite"></param>
        public static void RemoveSprite(Sprite sprite)
        {
            sprites.Remove(sprite);
        }


        /// <summary>
        /// Update all sprite animations
        /// </summary>
        /// <param name="gameTime"></param>
        public static void Update(GameTime gameTime)
        {
            foreach (Sprite s in sprites)
                s.Update(gameTime);
        }
    }


    public class Sprite
    {
        /// <summary>
        /// The SpriteSheet this sprite is based off of
        /// </summary>
        private SpriteSheet type;

        public bool HasAnimation(string anim)
        {
            return type.HasAnimation(anim);
        }

        /// <summary>
        /// The width of this sprite
        /// </summary>
        public int Width { get { return type.Width(animation, frame); } }
        /// <summary>
        /// The height of this sprite
        /// </summary>
        public int Height { get { return type.Height(animation, frame); } }

        /// <summary>
        /// The position of the sprite on the screen
        /// </summary>
        public Vector2 position;

        /// <summary>
        /// The current animation frame of the sprite
        /// </summary>
        public int frame;

        /// <summary>
        /// The current animation being played
        /// </summary>
        public string animation { get; private set; }

        /// <summary>
        /// The amount of time left to show this frame
        /// </summary>
        private TimeSpan thisFrame;

        /// <summary>
        /// Set if the animation's length has been set manually
        /// </summary>
        private TimeSpan? animationLength;

        /// <summary>
        /// Is this sprite playing an animation?
        /// </summary>
        public bool isPlaying { get; private set; }

        /// <summary>
        /// Is the current animation being looped?
        /// </summary>
        public bool isLooping { get; private set; }


        private Vector2 rootSize;
        private Vector2 sizeChangeCompensation
        {
            get
            {
                return (rootSize - new Vector2(Width, Height)) / 2
                    + type.files[type.animations[animation].frames[frame].file].centeringOffset;
            }
        }

        public delegate void FrameChangedDelegate(int lastFrame, int newFrame);
        public event FrameChangedDelegate FrameChanged;


        /// <summary>
        /// Create a new instance of the given spritesheet
        /// </summary>
        /// <param name="type"></param>
        public Sprite(SpriteSheet type)
        {
            this.type = type;

            SetFrame(type.animations.Keys.ElementAt(0), 0);
            rootSize = new Vector2(Width, Height);
            Stop();

            SpriteManager.AddSprite(this);
        }


        /// <summary>
        /// Create a new instance of the given spritesheet
        /// </summary>
        /// <param name="type"></param>
        public Sprite(SpriteSheet type, Vector2 position)
        {
            this.type = type;

            this.position = position;

            SetFrame(type.animations.Keys.ElementAt(0), 0);
            Stop();

            SpriteManager.AddSprite(this);
        }


        /// <summary>
        /// Delete this sprite
        /// </summary>
        ~Sprite()
        {
            SpriteManager.RemoveSprite(this);
        }


        /// <summary>
        /// Stop the animation
        /// </summary>
        public void Stop()
        {
            thisFrame = TimeSpan.Zero;
            isPlaying = false;
            animationLength = null;
            isLooping = false;
        }


        /// <summary>
        /// Cause the current animation to loop
        /// </summary>
        public void Loop()
        {
            isLooping = true;
        }


        /// <summary>
        /// Progress this sprite's animation as needed.
        /// </summary>
        /// <param name="gameTime"></param>
        public void Update(GameTime gameTime)
        {
            //No change
            if (isPlaying)
            {
                if (thisFrame > gameTime.ElapsedGameTime)
                {
                    thisFrame -= gameTime.ElapsedGameTime;
                    return;
                }
                else
                {
                    int oldFrame = frame;

                    //Next frame
                    frame++;

                    //If this is past last frame
                    if (frame >= type.animations[animation].frames.Count)
                    {
                        if (isLooping)
                            frame = 0;
                        else
                        {
                            frame = type.animations[animation].frames.Count - 1;
                            Stop();
                            return;
                        }
                    }

                    if (FrameChanged != null
                        && oldFrame != frame)
                        FrameChanged(oldFrame, frame);

                    //Set time
                    thisFrame = type.animations[animation].frames[frame].length - (gameTime.ElapsedGameTime - thisFrame);
                }
            }
        }


        /// <summary>
        /// Set this sprite's frame manually
        /// </summary>
        /// <param name="frame"></param>
        public void SetFrame(string animation, int frame)
        {
            this.animation = animation;
            this.frame = frame;
            Stop();
        }


        /// <summary>
        /// Tell the sprite to play the given animation
        /// </summary>
        /// <param name="animation"></param>
        public void PlayAnimation(string animation)
        {
            if (!type.animations.ContainsKey(animation)) return;
            Stop();
            isPlaying = true;
            this.animation = animation;
            animationLength = null;
            frame = 0;
            thisFrame = type.animations[animation].frames[frame].length;
        }


        /// <summary>
        /// Tell the sprite to play the given animation
        /// </summary>
        /// <param name="animation"></param>
        public void PlayAnimation(string animation, TimeSpan length)
        {
            Stop();
            this.animation = animation;
            animationLength = length;
            frame = 0;
            isPlaying = true;
            thisFrame = type.animations[animation].frames[frame].length;
        }


        /// <summary>
        /// Draw this sprite to the screen
        /// </summary>
        /// <param name="batch"></param>
        /// <param name="position"></param>
        public void Draw(SpriteBatch batch, Vector2 position, Color color, float rotation, Vector2 origin, float scale)
        {
            type.Draw(
                batch,
                position + sizeChangeCompensation * scale,
                type.animations[animation].frames[frame].file,
                type.animations[animation].frames[frame].id,
                color,
                rotation,
                origin,
                scale,
                SpriteEffects.None);
        }


        /// <summary>
        /// Draw this sprite to the screen
        /// </summary>
        /// <param name="batch"></param>
        /// <param name="position"></param>
        public void Draw(SpriteBatch batch, Vector2 position, Color color, float rotation, Vector2 origin, float scale, SpriteEffects effects)
        {
            type.Draw(
                batch,
                position + sizeChangeCompensation * scale,
                type.animations[animation].frames[frame].file,
                type.animations[animation].frames[frame].id,
                color,
                rotation,
                origin,
                scale,
                effects);
        }


        /// <summary>
        /// Draw this sprite to the screen
        /// </summary>
        /// <param name="batch"></param>
        /// <param name="position"></param>
        public void Draw(SpriteBatch batch, Color color, float rotation, Vector2 origin, float scale)
        {
            type.Draw(
                batch,
                position + sizeChangeCompensation * scale,
                type.animations[animation].frames[frame].file,
                type.animations[animation].frames[frame].id,
                color,
                rotation,
                origin,
                scale,
                SpriteEffects.None);
        }


        /// <summary>
        /// Draw this sprite's normals to the screen
        /// </summary>
        /// <param name="batch"></param>
        /// <param name="position"></param>
        public void DrawNormal(SpriteBatch batch, Color color, float rotation, Vector2 origin, float scale)
        {
            type.DrawNormal(
                batch,
                position + sizeChangeCompensation * scale,
                type.animations[animation].frames[frame].file,
                type.animations[animation].frames[frame].id,
                color,
                rotation,
                origin,
                scale,
                SpriteEffects.None);
        }


        /// <summary>
        /// Returns the location of the first pixel of the given color
        /// </summary>
        /// <param name="color"></param>
        /// <returns></returns>
        public Vector2? FindPixel(Color color)
        {
            return type.FindPixel(animation, frame, color);
        }
    }


    /// <summary>
    /// This class represents an image divided into a series of
    /// frames that can be drawn as an animation.
    /// </summary>
    public class SpriteSheet
    {
        /// <summary>
        /// The width of one frame of this spritesheet.
        /// </summary>
        public int Width(string animation, int frame)
        {
            return files[animations[animation].frames[frame].file].width;
        }

        /// <summary>
        /// The width of one frame of this spritesheet.
        /// </summary>
        public int Height(string animation, int frame)
        {
            return files[animations[animation].frames[frame].file].height;
        }


        /// <summary>
        /// The animations defined in this spritesheet.
        /// </summary>
        public Dictionary<string, SpriteAnimation> animations;

        /// <summary>
        /// Does this spritesheet contain an animation by the given name?
        /// </summary>
        /// <param name="anim"></param>
        /// <returns></returns>
        public bool HasAnimation(string anim)
        {
            foreach (string s in animations.Keys)
                if (s == anim) return true;
            return false;
        }

        /// <summary>
        /// The files used in this spritesheet.
        /// </summary>
        public new Dictionary<string, SpriteFile> files;

        //TODO: Can draw frames from files with multiple rows

        /// <summary>
        /// Construct a new spritesheet object using the given image and data.
        /// </summary>
        /// <param name="filename"></param>
        /// <param name="tileSize">The size of one tile on the image.</param>
        public SpriteSheet()
        {
            animations = new Dictionary<string, SpriteAnimation>();
            files = new Dictionary<string, SpriteFile>();
        }


        /// <summary>
        /// Returns the location of the first pixel of the given color
        /// </summary>
        /// <param name="animation"></param>
        /// <param name="frame"></param>
        /// <param name="color"></param>
        /// <returns></returns>
        public Vector2? FindPixel(string animation, int frame, Color color)
        {
            Color[] data = new Color[Width(animation, frame) * Height(animation, frame)];
            GetAnimationFrame(animation, frame).GetData<Color>(data);
            for (int c = 0; c < data.Length; c++)
                if (data[c] == color)
                    return new Vector2(c % Width(animation, frame),
                        (int)Math.Floor(c / (float)Width(animation, frame)));
            return null;
        }


        /// <summary>
        /// Get the image for the given frame
        /// </summary>
        /// <param name="file"></param>
        /// <param name="frame"></param>
        /// <returns></returns>
        public Texture2D GetAnimationFrame(string animation, int aframe)
        {
            SpriteFile file = files[animations[animation].frames[aframe].file];
            int frame = animations[animation].frames[aframe].id;
            return file.GetFrame(frame);
        }


        /// <summary>
        /// Counts the number of frames in the given animation
        /// </summary>
        /// <param name="animation"></param>
        /// <returns></returns>
        public int FrameCount(string animation)
        {
            return animations[animation].frames.Count;
        }


        /// <summary>
        /// Load the spritesheet images
        /// </summary>
        public void LoadContent(ContentManager content)
        {
            foreach (SpriteFile s in files.Values)
                s.LoadContent(content);
        }


        /// <summary>
        /// Draws a frame of this spritesheet to the screen.
        /// </summary>
        public void Draw(SpriteBatch batch, Vector2 position, string file, int frame, Color color, float rotation, Vector2 origin, float scale, SpriteEffects effects)
        {
            //Check that frame is valid
            if (frame * files[file].width >= files[file].totalwidth || frame < 0)
                throw new ArgumentOutOfRangeException("frame");

            //Draw
            batch.Draw(
                files[file].GetFrame(frame),
                position,
                null,
                color,
                rotation,
                origin,
                scale,
                effects,
                0.9f);
        }


        public void DrawNormal(SpriteBatch batch, Vector2 position, string file, int frame, Color color, float rotation, Vector2 origin, float scale, SpriteEffects effects)
        {
            //Check that frame is valid
            if (frame * files[file].width >= files[file].totalwidth || frame < 0
                || files[file].GetFrameNormalMap(frame) == null)
                throw new ArgumentOutOfRangeException("frame");

            Texture2D useimg = files[file].GetFrameNormalMap(frame);

            //Gotta flip reds or greens if sprite is flipped
            if (effects == SpriteEffects.FlipHorizontally)
            {
                Color[] data = new Color[useimg.Width * useimg.Height];
                useimg.GetData<Color>(data);
                for (int c = 0; c < data.Length; c++)
                    data[c].R = (byte)(255 - data[c].R);
                useimg.SetData<Color>(data);
            }
            else if (effects == SpriteEffects.FlipVertically)
            {
                Color[] data = new Color[useimg.Width * useimg.Height];
                useimg.GetData<Color>(data);
                for (int c = 0; c < data.Length; c++)
                    data[c].G = (byte)(255 - data[c].G);
                useimg.SetData<Color>(data);
            }

            //Draw
            batch.Draw(
                useimg,
                position,
                null,
                color,
                rotation,
                origin,
                scale,
                effects,
                0.9f);
        }


        /// <summary>
        /// Add a new animation to this spritesheet.
        /// </summary>
        /// <param name="animation"></param>
        public void AddAnimation(SpriteAnimation animation)
        {
            animations.Add(animation.name, animation);
        }


        /// <summary>
        /// Adds a new file to this spritesheet.
        /// </summary>
        /// <param name="filename"></param>
        public void AddFile(SpriteFile file)
        {
            files.Add(file.name, file);
        }
    }


    public class SpriteFile
    {
        public Texture2D image;

        public Texture2D normal;

        public string name;

        public string normalName;

        /// <summary>
        /// Get the width of a frame in this file
        /// </summary>
        /// <param name="frame"></param>
        /// <returns></returns>
        public int width;

        /// <summary>
        /// Get the height of a frame in this file
        /// </summary>
        /// <param name="frame"></param>
        /// <returns></returns>
        public int height;

        public int totalwidth { get { return image.Width; } }

        public int totalheight { get { return image.Height; } }

        public Vector2 centeringOffset;

        public SpriteFile(string name, string normal, int width, int height, int offx, int offy)
        {
            this.name = name;
            this.normalName = normal;
            this.width = width;
            this.height = height;
            centeringOffset = new Vector2(offx, offy);
        }

        public void LoadContent(ContentManager content)
        {
            image = content.Load<Texture2D>(name);
            if (normalName != "")
                normal = content.Load<Texture2D>(normalName);
        }

        public Texture2D GetFrame(int frame)
        {
            return GetFrame(image, frame);
        }

        public Texture2D GetFrameNormalMap(int frame)
        {
            return GetFrame(normal, frame);
        }

        private Texture2D GetFrame(Texture2D image, int frame)
        {
            Color[] data = new Color[width * height];
            image.GetData<Color>(
                0,
                new Rectangle(
                    frame * width,
                    0,
                    width,
                    height),
                data,
                0, data.Length);
            Texture2D ret = new Texture2D(
                image.GraphicsDevice,
                width,
                height);
            ret.SetData<Color>(data);
            return ret;
        }
    }


    /// <summary>
    /// Represents one animation in the spritesheet.
    /// </summary>
    public class SpriteAnimation
    {
        /// <summary>
        /// The name of this animation.
        /// </summary>
        public string name;

        /// <summary>
        /// The frames that make up this animation.
        /// </summary>
        public List<Frame> frames;

        /// <summary>
        /// The overall length of the animation.
        /// </summary>
        public TimeSpan Length
        {
            get
            {
                TimeSpan sum = TimeSpan.Zero;
                foreach (Frame frame in frames)
                    sum += frame.length;
                return sum;
            }
        }

        /// <summary>
        /// Construct a new animation object.
        /// </summary>
        public SpriteAnimation(string name)
        {
            this.name = name;

            frames = new List<Frame>();
        }

        /// <summary>
        /// Add a new frame to this animation.
        /// </summary>
        /// <param name="frame"></param>
        public void AddFrame(Frame frame)
        {
            frames.Add(frame);
        }
    }


    /// <summary>
    /// Represents one frame in an animation.
    /// </summary>
    public struct Frame
    {
        /// <summary>
        /// The location of this frame on the image.
        /// </summary>
        public int id;

        /// <summary>
        /// The length of this frame.
        /// </summary>
        public TimeSpan length;

        /// <summary>
        /// The file this frame comes from
        /// </summary>
        public string file;

        /// <summary>
        /// Construct a new, empty Frame object.
        /// </summary>
        public Frame(int id, int length, string file)
        {
            this.id = id;
            this.length = new TimeSpan(0, 0, 0, 0, length);
            this.file = file;
        }
    }


    /// <summary>
    /// This class reads binary spritesheet content files.
    /// </summary>
    public class SpriteSheetReader : ContentTypeReader<SpriteSheet>
    {
        protected override SpriteSheet Read(ContentReader input, SpriteSheet existingInstance)
        {
            SpriteSheet ret = new SpriteSheet();

            //Read files
            int fileCount = input.ReadInt32();
            for (int file = 0; file < fileCount; file++)
            {
                ret.AddFile(new SpriteFile(
                    input.ReadString(),
                    input.ReadString(),
                    input.ReadInt32(),
                    input.ReadInt32(),
                    input.ReadInt32(),
                    input.ReadInt32()));
            }

            //Read animations
            int animCount = input.ReadInt32();
            for (int anim = 0; anim < animCount; anim++)
            {
                SpriteAnimation building = new SpriteAnimation(input.ReadString());

                //Read frames
                int frameCount = input.ReadInt32();
                for (int frame = 0; frame < frameCount; frame++)
                {
                    building.AddFrame(new Frame(
                        input.ReadInt32(),
                        input.ReadInt32(),
                        input.ReadString()));
                }

                ret.AddAnimation(building);
            }

            return ret;
        }
    }
}
